from typing import Annotated, Optional
from uuid import UUID

from duckdb import DuckDBPyConnection

from arthur_common.aggregations.aggregator import NumericAggregationFunction
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    NumericMetric,
)
from arthur_common.models.schema_definitions import (
    SEGMENTATION_ALLOWED_COLUMN_TYPES,
    DType,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
    MetricMultipleColumnParameterAnnotation,
    ScalarType,
    ScopeSchemaTag,
)
from arthur_common.tools.duckdb_data_loader import (
    escape_str_literal,
    unescape_identifier,
)


class CategoricalCountAggregationFunction(NumericAggregationFunction):
    METRIC_NAME = "categorical_count"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-00000000000c")

    @staticmethod
    def display_name() -> str:
        return "Category Count"

    @staticmethod
    def description() -> str:
        return "Metric that counts the number of discrete values of each category in a string column. Creates a separate dimension for each category and the values are the count of occurrences of that category in the time window."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=CategoricalCountAggregationFunction.METRIC_NAME,
                description=CategoricalCountAggregationFunction.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing some categorical data.",
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        categorical_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.STRING),
                    ScalarType(dtype=DType.INT),
                ],
                tag_hints=[ScopeSchemaTag.CATEGORICAL],
                friendly_name="Categorical Column",
                description="A column containing categorical values to count.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        """Executed SQL with no segmentation columns:
            select time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts, \
                count(*) as count, \
                {categorical_col} as category, \
                {categorical_col_name_unescaped} as column_name \
                from {dataset.dataset_table_name} \
                where ts is not null \
                group by ts, category
        """
        segmentation_cols = [] if not segmentation_cols else segmentation_cols
        categorical_col_name_unescaped = escape_str_literal(
            unescape_identifier(categorical_col),
        )

        # build query components with segmentation columns
        all_select_clause_cols = [
            f"time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts",
            f"count(*) as count",
            f"{categorical_col} as category",
            f"{categorical_col_name_unescaped} as column_name",
        ] + segmentation_cols
        all_group_by_cols = ["ts", "category"] + segmentation_cols
        extra_dims = ["column_name", "category"]

        # build query
        count_query = f"""
            select {", ".join(all_select_clause_cols)}
            from {dataset.dataset_table_name}
            where ts is not null
            group by {", ".join(all_group_by_cols)}
        """

        results = ddb_conn.sql(count_query).df()

        series = self.group_query_results_to_numeric_metrics(
            results,
            "count",
            segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]



import duckdb

from arthur_common.config.config import Config
from arthur_common.models.schema_definitions import SEGMENTATION_ALLOWED_DTYPES, DType


def is_column_possible_segmentation(
    conn: duckdb.DuckDBPyConnection,
    table: str,
    column_name: str,
    column_dtype: DType,
) -> bool:
    """Returns whether column fits segmentation criteria:
    1. Has fewer than SEGMENTATION_COL_UNIQUE_VALUE_LIMIT unique values.
    2. Has an allowed DType.

    PreReq: Table with column should already be loaded in DuckDB
    column_name already has DuckDB escape identifier for the query syntax
    """
    segmentation_col_unique_val_limit = Config.segmentation_col_unique_values_limit()
    if column_dtype not in SEGMENTATION_ALLOWED_DTYPES:
        return False

    # check column for unique value count - count distinct values in this column
    distinct_count_query = f"""
        SELECT COUNT(DISTINCT {column_name}) as distinct_count
        FROM {table}
    """
    result = conn.sql(distinct_count_query).fetchone()
    distinct_count = result[0] if result else 0

    return distinct_count < segmentation_col_unique_val_limit



from enum import IntEnum, StrEnum

from arthur_common.models.constants import (
    DEFAULT_RULE_ADMIN,
    ORG_ADMIN,
    ORG_AUDITOR,
    TASK_ADMIN,
    VALIDATION_USER,
)


class BaseEnum(StrEnum):
    @classmethod
    def values(self) -> list[str]:
        values: list[str] = [e for e in self]
        return values

    def __str__(self) -> str:
        return str(self.value)


class APIKeysRolesEnum(BaseEnum):
    DEFAULT_RULE_ADMIN = DEFAULT_RULE_ADMIN
    TASK_ADMIN = TASK_ADMIN
    VALIDATION_USER = VALIDATION_USER
    ORG_AUDITOR = ORG_AUDITOR
    ORG_ADMIN = ORG_ADMIN


class InferenceFeedbackTarget(BaseEnum):
    CONTEXT = "context"
    RESPONSE_RESULTS = "response_results"
    PROMPT_RESULTS = "prompt_results"


class MetricType(BaseEnum):
    QUERY_RELEVANCE = "QueryRelevance"
    RESPONSE_RELEVANCE = "ResponseRelevance"
    TOOL_SELECTION = "ToolSelection"


class ModelProblemType(BaseEnum):
    REGRESSION = "regression"
    BINARY_CLASSIFICATION = "binary_classification"
    ARTHUR_SHIELD = "arthur_shield"
    CUSTOM = "custom"
    MULTICLASS_CLASSIFICATION = "multiclass_classification"
    AGENTIC_TRACE = "agentic_trace"


class SnowflakeConnectorAuthenticatorMethods(BaseEnum):
    SNOWFLAKE_KEY_PAIR = "snowflake_key_pair"
    SNOWFLAKE_PASSWORD = "snowflake_password"


# Using version from arthur-engine, which has str and enum type inheritance.
# Note: These string values are not arbitrary and map to Presidio entity types: https://microsoft.github.io/presidio/supported_entities/
class PIIEntityTypes(BaseEnum):
    CREDIT_CARD = "CREDIT_CARD"
    CRYPTO = "CRYPTO"
    DATE_TIME = "DATE_TIME"
    EMAIL_ADDRESS = "EMAIL_ADDRESS"
    IBAN_CODE = "IBAN_CODE"
    IP_ADDRESS = "IP_ADDRESS"
    NRP = "NRP"
    LOCATION = "LOCATION"
    PERSON = "PERSON"
    PHONE_NUMBER = "PHONE_NUMBER"
    MEDICAL_LICENSE = "MEDICAL_LICENSE"
    URL = "URL"
    US_BANK_NUMBER = "US_BANK_NUMBER"
    US_DRIVER_LICENSE = "US_DRIVER_LICENSE"
    US_ITIN = "US_ITIN"
    US_PASSPORT = "US_PASSPORT"
    US_SSN = "US_SSN"

    @classmethod
    def to_string(cls) -> str:
        return ",".join(member.value for member in cls)


class PaginationSortMethod(BaseEnum):
    ASCENDING = "asc"
    DESCENDING = "desc"


class RuleResultEnum(BaseEnum):
    PASS = "Pass"
    FAIL = "Fail"
    SKIPPED = "Skipped"
    UNAVAILABLE = "Unavailable"
    PARTIALLY_UNAVAILABLE = "Partially Unavailable"
    MODEL_NOT_AVAILABLE = "Model Not Available"


class RuleScope(BaseEnum):
    DEFAULT = "default"
    TASK = "task"


class RuleType(BaseEnum):
    KEYWORD = "KeywordRule"
    MODEL_HALLUCINATION_V2 = "ModelHallucinationRuleV2"
    MODEL_SENSITIVE_DATA = "ModelSensitiveDataRule"
    PII_DATA = "PIIDataRule"
    PROMPT_INJECTION = "PromptInjectionRule"
    REGEX = "RegexRule"
    TOXICITY = "ToxicityRule"


class TaskType(BaseEnum):
    TRADITIONAL = "traditional"
    AGENTIC = "agentic"


class TokenUsageScope(BaseEnum):
    RULE_TYPE = "rule_type"
    TASK = "task"


class ToolClassEnum(IntEnum):
    INCORRECT = 0
    CORRECT = 1
    NA = 2

    def __str__(self) -> str:
        return str(self.value)


class ToxicityViolationType(BaseEnum):
    BENIGN = "benign"
    HARMFUL_REQUEST = "harmful_request"
    TOXIC_CONTENT = "toxic_content"
    PROFANITY = "profanity"
    UNKNOWN = "unknown"


# If you added values here, did you update permission_mappings.py in arthur-engine?
class UserPermissionAction(BaseEnum):
    CREATE = "create"
    READ = "read"


# If you added values here, did you update permission_mappings.py in arthur-engine?
class UserPermissionResource(BaseEnum):
    PROMPTS = "prompts"
    RESPONSES = "responses"
    RULES = "rules"
    TASKS = "tasks"


class ComparisonOperatorEnum(BaseEnum):
    EQUAL = "eq"
    GREATER_THAN = "gt"
    GREATER_THAN_OR_EQUAL = "gte"
    LESS_THAN = "lt"
    LESS_THAN_OR_EQUAL = "lte"


class StatusCodeEnum(BaseEnum):
    OK = "Ok"
    ERROR = "Error"
    UNSET = "Unset"


class ContinuousEvalRunStatus(BaseEnum):
    PENDING = "pending"
    PASSED = "passed"
    RUNNING = "running"
    FAILED = "failed"
    SKIPPED = "skipped"
    ERROR = "error"


class AgentPollingStatus(BaseEnum):
    PENDING = "pending"
    IDLE = "idle"
    RUNNING = "running"
    ERROR = "error"


class AgenticAnnotationType(BaseEnum):
    HUMAN = "human"
    CONTINUOUS_EVAL = "continuous_eval"


class RegisteredAgentProvider(BaseEnum):
    GCP = "gcp"



from datetime import datetime
from typing import Any, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from arthur_common.models.common_schemas import (
    AuthUserRole,
    ExamplesConfig,
    KeywordsConfig,
    PIIConfig,
    RegexConfig,
    ToxicityConfig,
    VariableTemplateValue,
)
from arthur_common.models.enums import (
    AgenticAnnotationType,
    ContinuousEvalRunStatus,
    InferenceFeedbackTarget,
    MetricType,
    PIIEntityTypes,
    RegisteredAgentProvider,
    RuleResultEnum,
    RuleScope,
    RuleType,
    ToxicityViolationType,
)


class GCPAgentMetadataResponse(BaseModel):
    """GCP-specific agent metadata for responses."""

    project_id: str = Field(description="Project ID of the agent.")
    region: str = Field(description="Region of the agent.")
    resource_id: str = Field(description="Resource ID of the agent.")


class AgentMetadataResponse(BaseModel):
    """Agent metadata for responses."""

    provider: RegisteredAgentProvider = Field(
        description="Provider of the registered agent.",
    )
    gcp_metadata: Optional[GCPAgentMetadataResponse] = Field(
        description="Metadata for the agent.",
        default=None,
    )

    class Config:
        use_enum_values = True


class HTTPError(BaseModel):
    detail: str

    model_config = ConfigDict(
        json_schema_extra={
            "example": {"detail": "HTTPException raised."},
        },
    )


class RuleResponse(BaseModel):
    id: str = Field(description="ID of the Rule")
    name: str = Field(description="Name of the Rule")
    type: RuleType = Field(description="Type of Rule")
    apply_to_prompt: bool = Field(description="Rule applies to prompt")
    apply_to_response: bool = Field(description="Rule applies to response")
    enabled: Optional[bool] = Field(
        description="Rule is enabled for the task",
        default=None,
    )
    scope: RuleScope = Field(
        description="Scope of the rule. The rule can be set at default level or task level.",
    )
    # UNIX millis format
    created_at: int = Field(
        description="Time the rule was created in unix milliseconds",
    )
    updated_at: int = Field(
        description="Time the rule was updated in unix milliseconds",
    )
    config: (
        KeywordsConfig
        | RegexConfig
        | ExamplesConfig
        | ToxicityConfig
        | PIIConfig
        | None
    ) = Field(description="Config of the rule", default=None)


class HallucinationClaimResponse(BaseModel):
    claim: str
    valid: bool
    reason: str
    order_number: Optional[int] = Field(
        default=-1,
        description="This field is a helper for ordering the claims",
    )


class PIIEntitySpanResponse(BaseModel):
    entity: PIIEntityTypes
    span: str = Field(
        description="The subtext within the input string that was identified as PII.",
    )
    # Only optional to keep reverse compatibility with old inferences
    confidence: Optional[float] = Field(
        description="Float value representing the confidence score of a given PII identification.",
        default=None,
    )


class KeywordSpanResponse(BaseModel):
    keyword: str = Field(
        description="The keyword from the rule that matched within the input string.",
    )


class RegexSpanResponse(BaseModel):
    matching_text: str = Field(
        description="The subtext within the input string that matched the regex rule.",
    )
    pattern: Optional[str] = Field(
        description="Pattern that yielded the match.",
        default=None,
    )


class BaseDetailsResponse(BaseModel):
    score: Optional[bool] = None
    message: Optional[str] = None


class HallucinationDetailsResponse(BaseDetailsResponse):
    claims: list[HallucinationClaimResponse]


class PIIDetailsResponse(BaseDetailsResponse):
    pii_entities: list[PIIEntitySpanResponse]


class ToxicityDetailsResponse(BaseDetailsResponse):
    toxicity_score: Optional[float] = None
    toxicity_violation_type: ToxicityViolationType

    model_config = ConfigDict(extra="forbid")


class KeywordDetailsResponse(BaseDetailsResponse):
    keyword_matches: list[KeywordSpanResponse] = Field(
        [],
        description="Each keyword in this list corresponds to a keyword that was both configured in the rule that was "
        "run and found in the input text.",
    )

    model_config = ConfigDict(extra="forbid")


class RegexDetailsResponse(BaseDetailsResponse):
    regex_matches: list[RegexSpanResponse] = Field(
        [],
        description="Each string in this list corresponds to a matching span from the input text that matches the "
        "configured regex rule.",
    )

    model_config = ConfigDict(extra="forbid")


class ExternalRuleResult(BaseModel):
    id: str = Field(description=" ID of the rule")
    name: str = Field(description="Name of the rule")
    rule_type: RuleType = Field(description="Type of the rule")
    scope: RuleScope = Field(
        description="Scope of the rule. The rule can be set at default level or task level.",
    )
    result: RuleResultEnum = Field(description="Result if the rule")
    latency_ms: int = Field(description="Duration in millisesconds of rule execution")

    # The super class (BaseDetailsResponse) must come last in this ordering otherwise the fastapi serializer will pick
    # it for the less specific types and you'll waste time figuring out why type1 is being serialized as type2
    # https://github.com/tiangolo/fastapi/issues/2783#issuecomment-776662347
    details: Optional[
        Union[
            KeywordDetailsResponse,
            RegexDetailsResponse,
            HallucinationDetailsResponse,
            PIIDetailsResponse,
            ToxicityDetailsResponse,
            BaseDetailsResponse,
        ]
    ] = Field(description="Details of the rule output", default=None)

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "id": "90f18c69-d793-4913-9bde-a0c7f3643de0",
                "name": "PII Rule",
                "result": "Pass",
            },
        },
    )


class ValidationResult(BaseModel):
    inference_id: Optional[str] = Field(description="ID of the inference", default=None)
    rule_results: Optional[List[ExternalRuleResult]] = Field(
        description="List of rule results",
        default=None,
    )
    user_id: Optional[str] = Field(
        description="The user ID this prompt belongs to",
        default=None,
    )
    model_name: Optional[str] = Field(
        description="The model name and version used for this validation (e.g., 'gpt-4', 'gpt-3.5-turbo', 'claude-3-opus', 'gemini-pro').",
        default=None,
    )
    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "inference_id": "4dd1fae1-34b9-4aec-8abe-fe7bf12af31d",
                "rule_results": [
                    {
                        "id": "90f18c69-d793-4913-9bde-a0c7f3643de0",
                        "name": "PII Check",
                        "result": "Pass",
                    },
                    {
                        "id": "946c4a44-b367-4229-84d4-1a8e461cb132",
                        "name": "Sensitive Data Check",
                        "result": "Pass",
                    },
                ],
            },
        },
    )


class ExternalInferencePrompt(BaseModel):
    id: str
    inference_id: str
    result: RuleResultEnum
    created_at: int
    updated_at: int
    message: str
    prompt_rule_results: List[ExternalRuleResult]
    tokens: int | None = None


class ExternalInferenceResponse(BaseModel):
    id: str
    inference_id: str
    result: RuleResultEnum
    created_at: int
    updated_at: int
    message: str
    context: Optional[str] = None
    response_rule_results: List[ExternalRuleResult]
    tokens: int | None = None
    model_name: Optional[str] = Field(
        description="The model name and version used for this response (e.g., 'gpt-4', 'gpt-3.5-turbo', 'claude-3-opus', 'gemini-pro').",
        default=None,
    )


class InferenceFeedbackResponse(BaseModel):
    id: str
    inference_id: str
    target: InferenceFeedbackTarget
    score: int
    reason: Optional[str] = None
    user_id: Optional[str] = None
    created_at: datetime
    updated_at: datetime


class QueryFeedbackResponse(BaseModel):
    feedback: list[InferenceFeedbackResponse] = Field(
        description="List of inferences matching the search filters. Length is less than or equal to page_size parameter",
    )
    page: int = Field(description="The current page number")
    page_size: int = Field(description="The number of feedback items per page")
    total_pages: int = Field(description="The total number of pages")
    total_count: int = Field(
        description="The total number of feedback items matching the query parameters",
    )

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "feedback": [
                    {
                        "id": "90f18c69-d793-4913-9bde-a0c7f3643de0",
                        "inference_id": "81437d71-9557-4611-981b-9283d1c98643",
                        "target": "context",
                        "score": "0",
                        "reason": "good reason",
                        "user_id": "user_1",
                        "created_at": "2024-06-06T06:37:46.123-04:00",
                        "updated_at": "2024-06-06T06:37:46.123-04:00",
                    },
                    {
                        "id": "248381c2-543b-4de0-98cd-d7511fee6241",
                        "inference_id": "bcbc7ca0-4cfc-4f67-9cf8-26cb2291ba33",
                        "target": "response_results",
                        "score": "1",
                        "reason": "some reason",
                        "user_id": "user_2",
                        "created_at": "2023-05-05T05:26:35.987-04:00",
                        "updated_at": "2023-05-05T05:26:35.987-04:00",
                    },
                ],
                "page": 1,
                "page_size": 10,
                "total_pages": 1,
                "total_count": 2,
            },
        },
    )


class ExternalInference(BaseModel):
    id: str
    result: RuleResultEnum
    created_at: int
    updated_at: int
    task_id: Optional[str] = None
    task_name: str | None = None
    conversation_id: Optional[str] = None
    inference_prompt: ExternalInferencePrompt
    inference_response: Optional[ExternalInferenceResponse] = None
    inference_feedback: List[InferenceFeedbackResponse]
    user_id: str | None = None
    model_name: Optional[str] = Field(
        description="The model name and version used for this inference (e.g., 'gpt-4', 'gpt-3.5-turbo', 'claude-3-opus', 'gemini-pro').",
        default=None,
    )


class QueryInferencesResponse(BaseModel):
    count: int = Field(
        description="The total number of inferences matching the query parameters",
    )
    inferences: list[ExternalInference] = Field(
        description="List of inferences matching the search filters. Length is less than or equal to page_size parameter",
    )

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "count": 1,
                "inferences": [
                    {
                        "id": "957df309-c907-4b77-abe5-15dd00c081f7",
                        "result": "Pass",
                        "created_at": 1723204737120,
                        "updated_at": 1723204787050,
                        "task_id": "957df309-c907-4b77-abe5-15dd00c081f8",
                        "task_name": "My task name",
                        "conversation_id": "957df309-c907-4b77-abe5-15dd00c08112",
                        "inference_prompt": {
                            "id": "834f7ebd-cd6b-4691-9473-8bc350f8922c",
                            "inference_id": "957df309-c907-4b77-abe5-15dd00c081f7",
                            "result": "Pass",
                            "created_at": 1723204737121,
                            "updated_at": 1723204737121,
                            "message": "How many stars are in the solar system?",
                            "prompt_rule_results": [
                                {
                                    "id": "bc599a56-2e31-4cb7-910d-9e5ed6455db2",
                                    "name": "My_PII_Rule",
                                    "rule_type": "PIIDataRule",
                                    "scope": "default",
                                    "result": "Pass",
                                    "latency_ms": 73,
                                    "details": None,
                                },
                            ],
                            "tokens": 100,
                        },
                        "inference_response": {
                            "id": "ec765a75-1479-4938-8e1c-6334b7deb8ce",
                            "inference_id": "957df309-c907-4b77-abe5-15dd00c081f7",
                            "result": "Pass",
                            "created_at": 1723204786599,
                            "updated_at": 1723204786599,
                            "message": "There is one star in solar system.",
                            "context": "Solar system contains one star.",
                            "response_rule_results": [
                                {
                                    "id": "a45267c5-96d9-4de2-a871-debf2c8fdb86",
                                    "name": "My_another_PII_Rule",
                                    "rule_type": "PIIDataRule",
                                    "scope": "default",
                                    "result": "Pass",
                                    "latency_ms": 107,
                                    "details": None,
                                },
                                {
                                    "id": "92b7b46e-eaf2-4226-82d4-be12ceb3e4b7",
                                    "name": "My_Hallucination_Rule",
                                    "rule_type": "ModelHallucinationRuleV2",
                                    "scope": "default",
                                    "result": "Pass",
                                    "latency_ms": 700,
                                    "details": {
                                        "score": True,
                                        "message": "All claims were supported by the context!",
                                        "claims": [
                                            {
                                                "claim": "There is one star in solar system.",
                                                "valid": True,
                                                "reason": "No hallucination detected!",
                                                "order_number": 0,
                                            },
                                        ],
                                        "pii_results": [],
                                        "pii_entities": [],
                                        "toxicity_score": None,
                                    },
                                },
                            ],
                            "tokens": 100,
                        },
                        "inference_feedback": [
                            {
                                "id": "0d602e5c-4ae6-4fc9-a610-68a1d8928ad7",
                                "inference_id": "957df309-c907-4b77-abe5-15dd00c081f7",
                                "target": "context",
                                "score": 100,
                                "reason": "Perfect answer.",
                                "user_id": "957df309-2137-4b77-abe5-15dd00c081f8",
                                "created_at": "2024-08-09T12:08:34.847381",
                                "updated_at": "2024-08-09T12:08:34.847386",
                            },
                        ],
                        "user_id": "957df309-2137-4b77-abe5-15dd00c081f8",
                    },
                ],
            },
        },
    )


class MetricResponse(BaseModel):
    id: str = Field(description="ID of the Metric")
    name: str = Field(description="Name of the Metric")
    type: MetricType = Field(description="Type of the Metric")
    metric_metadata: str = Field(description="Metadata of the Metric")
    config: Optional[str] = Field(
        description="JSON-serialized configuration for the Metric",
        default=None,
    )
    created_at: datetime = Field(
        description="Time the Metric was created in unix milliseconds",
    )
    updated_at: datetime = Field(
        description="Time the Metric was updated in unix milliseconds",
    )
    enabled: Optional[bool] = Field(
        description="Whether the Metric is enabled",
        default=None,
    )


class TaskResponse(BaseModel):
    id: str = Field(description=" ID of the task")
    name: str = Field(description="Name of the task")
    created_at: int = Field(
        description="Time the task was created in unix milliseconds",
    )
    updated_at: int = Field(
        description="Time the task was created in unix milliseconds",
    )
    is_agentic: Optional[bool] = Field(
        description="Whether the task is agentic or not",
        default=None,
    )
    agent_metadata: Optional[AgentMetadataResponse] = Field(
        description="Metadata for registered agents.",
        default=None,
    )
    rules: List[RuleResponse] = Field(description="List of all the rules for the task.")
    metrics: Optional[List[MetricResponse]] = Field(
        description="List of all the metrics for the task.",
        default=None,
    )


class SearchTasksResponse(BaseModel):
    count: int = Field(description="The total number of tasks matching the parameters")
    tasks: list[TaskResponse] = Field(
        description="List of tasks matching the search filters. Length is less than or equal to page_size parameter",
    )


class SearchRulesResponse(BaseModel):
    count: int = Field(description="The total number of rules matching the parameters")
    rules: list[RuleResponse] = Field(
        description="List of rules matching the search filters. Length is less than or equal to page_size parameter",
    )


class FileUploadResult(BaseModel):
    id: str
    name: str
    type: str
    word_count: int
    success: bool


class ExternalDocument(BaseModel):
    id: str
    name: str
    type: str
    owner_id: str


class ChatDocumentContext(BaseModel):
    id: str
    seq_num: int
    context: str


class ChatResponse(BaseModel):
    inference_id: str = Field(description="ID of the inference sent to the chat")
    conversation_id: str = Field(description="ID of the conversation session")
    timestamp: int = Field(
        description="Time the inference was made in unix milliseconds",
    )
    retrieved_context: List[ChatDocumentContext] = Field(
        description="related sections of documents that were most relevant to the inference prompt. "
        "Formatted as a list of retrieved context chunks which include document name, seq num, and context.",
    )
    llm_response: str = Field(
        description="response from the LLM for the original user prompt",
    )
    prompt_results: List[ExternalRuleResult] = Field(
        description="list of rule results for the user prompt",
    )
    response_results: List[ExternalRuleResult] = Field(
        description="list of rule results for the llm response",
    )
    model_name: Optional[str] = Field(
        description="The model name and version used for this chat response (e.g., 'gpt-4', 'gpt-3.5-turbo', 'claude-3-opus', 'gemini-pro').",
        default=None,
    )


class TokenUsageCount(BaseModel):
    inference: int = Field(description="Number of inference tokens sent to Arthur.")
    eval_prompt: int = Field(
        description="Number of Prompt tokens incurred by Arthur rules.",
    )
    eval_completion: int = Field(
        description="Number of Completion tokens incurred by Arthur rules.",
    )
    user_input: int = Field(
        description="Number of user input tokens sent to Arthur. This field is deprecated and will be removed in the future. Use inference instead.",
        json_schema_extra={"deprecated": True},
    )
    prompt: int = Field(
        description="Number of Prompt tokens incurred by Arthur rules. This field is deprecated and will be removed in the future. Use eval_prompt instead.",
        json_schema_extra={"deprecated": True},
    )
    completion: int = Field(
        description="Number of Completion tokens incurred by Arthur rules. This field is deprecated and will be removed in the future. Use eval_completion instead.",
        json_schema_extra={"deprecated": True},
    )


class TokenUsageResponse(BaseModel):
    rule_type: Optional[str] = None
    task_id: Optional[str] = None
    count: TokenUsageCount


class ApiKeyResponse(BaseModel):
    id: str = Field(description="ID of the key")
    key: Optional[str] = Field(
        description="The generated GenAI Engine API key. The key is displayed on key creation request only.",
        default=None,
    )
    description: Optional[str] = Field(
        description="Description of the API key",
        default=None,
    )
    is_active: bool = Field(description="Status of the key.")
    created_at: datetime = Field(description="Creation time of the key")
    deactivated_at: Optional[datetime] = Field(
        description="Deactivation time of the key",
        default=None,
    )
    message: Optional[str] = Field(description="Optional Message", default=None)
    roles: list[str] = Field(
        description="Roles of the API key",
        default=[],
    )


class UserResponse(BaseModel):
    id: str
    email: str
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    roles: list[AuthUserRole]


class ConversationBaseResponse(BaseModel):
    id: str
    updated_at: datetime


class ConversationResponse(ConversationBaseResponse):
    inferences: list[ExternalInference]


class HealthResponse(BaseModel):
    message: str
    build_version: Optional[str] = None


class ChatDefaultTaskResponse(BaseModel):
    task_id: str


class MetricResultResponse(BaseModel):
    id: str = Field(description="ID of the metric result")
    metric_type: MetricType = Field(description="Type of the metric")
    details: Optional[str] = Field(
        description="JSON-serialized metric details",
        default=None,
    )
    prompt_tokens: int = Field(description="Number of prompt tokens used")
    completion_tokens: int = Field(description="Number of completion tokens used")
    latency_ms: int = Field(description="Latency in milliseconds")
    span_id: str = Field(description="ID of the span this result belongs to")
    metric_id: str = Field(description="ID of the metric that generated this result")
    created_at: datetime = Field(description="Time the result was created")
    updated_at: datetime = Field(description="Time the result was last updated")


class TokenCountCostSchema(BaseModel):
    """Base schema for responses that include token count and cost information.

    These fields represent LLM token usage and associated costs.
    None values indicate data is not available.
    """

    prompt_token_count: Optional[int] = Field(
        default=None,
        description="Number of prompt tokens",
    )
    completion_token_count: Optional[int] = Field(
        default=None,
        description="Number of completion tokens",
    )
    total_token_count: Optional[int] = Field(
        default=None,
        description="Total number of tokens",
    )
    prompt_token_cost: Optional[float] = Field(
        default=None,
        description="Cost of prompt tokens in USD",
    )
    completion_token_cost: Optional[float] = Field(
        default=None,
        description="Cost of completion tokens in USD",
    )
    total_token_cost: Optional[float] = Field(
        default=None,
        description="Total cost in USD",
    )


class SpanWithMetricsResponse(TokenCountCostSchema):
    id: str
    trace_id: str
    span_id: str
    parent_span_id: Optional[str] = None
    span_kind: Optional[str] = None
    span_name: Optional[str] = None
    start_time: datetime
    end_time: datetime
    task_id: Optional[str] = None
    session_id: Optional[str] = None
    user_id: Optional[str] = None
    status_code: str = Field(description="Status code for the span (Unset, Error, Ok)")
    created_at: datetime
    updated_at: datetime
    raw_data: dict[str, Any]
    # OpenInference standard input/output fields (computed on-demand from raw_data)
    input_content: Optional[str] = Field(
        None,
        description="Span input value from raw_data.attributes.input.value",
    )
    output_content: Optional[str] = Field(
        None,
        description="Span output value from raw_data.attributes.output.value",
    )
    metric_results: list[MetricResultResponse] = Field(
        description="List of metric results for this span",
        default=[],
    )


class NestedSpanWithMetricsResponse(TokenCountCostSchema):
    """Nested span response with children for building span trees"""

    id: str
    trace_id: str
    span_id: str
    parent_span_id: Optional[str] = None
    span_kind: Optional[str] = None
    span_name: Optional[str] = None
    start_time: datetime
    end_time: datetime
    task_id: Optional[str] = None
    session_id: Optional[str] = None
    user_id: Optional[str] = None
    status_code: str = Field(description="Status code for the span (Unset, Error, Ok)")
    created_at: datetime
    updated_at: datetime
    raw_data: dict[str, Any]
    # OpenInference standard input/output fields (computed on-demand from raw_data)
    input_content: Optional[str] = Field(
        None,
        description="Span input value from raw_data.attributes.input.value",
    )
    output_content: Optional[str] = Field(
        None,
        description="Span output value from raw_data.attributes.output.value",
    )
    metric_results: list[MetricResultResponse] = Field(
        description="List of metric results for this span",
        default=[],
    )
    children: list["NestedSpanWithMetricsResponse"] = Field(
        description="Child spans nested under this span",
        default=[],
    )


class AgenticAnnotationResponse(BaseModel):
    id: str = Field(description="ID of the annotation")
    annotation_type: AgenticAnnotationType = Field(description="Type of annotation")
    trace_id: str = Field(description="ID of the trace this annotation belongs to")
    continuous_eval_id: Optional[str] = Field(
        default=None,
        description="ID of the continuous eval this annotation belongs to",
    )
    continuous_eval_name: Optional[str] = Field(
        default=None,
        description="Name of the continuous eval this annotation belongs to",
    )
    eval_name: Optional[str] = Field(
        default=None,
        description="Name of the eval the continuous eval used when scoring",
    )
    eval_version: Optional[int] = Field(
        default=None,
        description="Version of the eval the continuous eval used when scoring",
    )
    annotation_score: Optional[int] = Field(
        default=None,
        description="Binary score for a positive or negative annotation.",
    )
    annotation_description: Optional[str] = Field(
        default=None,
        description="Description of the annotation.",
    )
    input_variables: Optional[List[VariableTemplateValue]] = Field(
        default=None,
        description="Input variables for the continuous eval",
    )
    run_status: Optional[ContinuousEvalRunStatus] = Field(
        default=None,
        description="Status of the continuous eval run",
    )
    cost: Optional[float] = Field(
        default=None,
        description="Cost of the continuous eval run",
    )
    created_at: datetime = Field(description="Time the annotation was created")
    updated_at: datetime = Field(description="Time the annotation was last updated")


class ListAgenticAnnotationsResponse(BaseModel):
    annotations: list[AgenticAnnotationResponse] = Field(
        description="List of annotations",
    )
    count: int = Field(description="Total number of annotations")


class TraceResponse(TokenCountCostSchema):
    """Response model for a single trace containing nested spans"""

    trace_id: str = Field(description="ID of the trace")
    start_time: datetime = Field(
        description="Start time of the earliest span in this trace",
    )
    end_time: datetime = Field(description="End time of the latest span in this trace")
    input_content: Optional[str] = Field(
        None,
        description="Root span input value from trace metadata",
    )
    output_content: Optional[str] = Field(
        None,
        description="Root span output value from trace metadata",
    )
    root_spans: list[NestedSpanWithMetricsResponse] = Field(
        description="Root spans (spans with no parent) in this trace, with children nested",
        default=[],
    )
    annotations: Optional[List[AgenticAnnotationResponse]] = Field(
        default=None,
        description="Annotations for this trace.",
    )


class QueryTracesWithMetricsResponse(BaseModel):
    """New response format that groups spans into traces with nested structure"""

    count: int = Field(
        description="The total number of spans matching the query parameters",
    )
    traces: list[TraceResponse] = Field(
        description="List of traces containing nested spans matching the search filters",
    )


class QuerySpansResponse(BaseModel):
    count: int = Field(
        description="The total number of spans matching the query parameters",
    )
    spans: list[SpanWithMetricsResponse] = Field(
        description="List of spans with metrics matching the search filters",
    )



from __future__ import annotations

import math
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator

from arthur_common.models.constants import (
    DEFAULT_PII_RULE_CONFIDENCE_SCORE_THRESHOLD,
    DEFAULT_TOXICITY_RULE_THRESHOLD,
    NEGATIVE_BLOOD_EXAMPLE,
)
from arthur_common.models.enums import (
    PaginationSortMethod,
    PIIEntityTypes,
    UserPermissionAction,
    UserPermissionResource,
)


class AuthUserRole(BaseModel):
    id: str | None = None
    name: str
    description: str
    composite: bool


class ExampleConfig(BaseModel):
    example: str = Field(description="Custom example for the sensitive data")
    result: bool = Field(
        description="Boolean value representing if the example passes or fails the the sensitive "
        "data rule ",
    )

    model_config = ConfigDict(
        json_schema_extra={
            "example": {"example": NEGATIVE_BLOOD_EXAMPLE, "result": True},
        },
    )


class ExamplesConfig(BaseModel):
    examples: List[ExampleConfig] = Field(
        description="List of all the examples for Sensitive Data Rule",
    )

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "examples": [
                    {"example": NEGATIVE_BLOOD_EXAMPLE, "result": True},
                    {
                        "example": "Most of the people have A positive blood group",
                        "result": False,
                    },
                ],
                "hint": "specific individual's blood type",
            },
        },
    )
    hint: Optional[str] = Field(
        description="Optional. Hint added to describe what Sensitive Data Rule should be checking for",
        default=None,
    )

    def to_dict(self) -> Dict[str, Any]:
        d = self.__dict__
        d["examples"] = [ex.__dict__ for ex in self.examples]
        d["hint"] = self.hint
        return d


class KeywordsConfig(BaseModel):
    keywords: List[str] = Field(description="List of Keywords")

    model_config = ConfigDict(
        json_schema_extra={
            "example": {"keywords": ["Blocked_Keyword_1", "Blocked_Keyword_2"]},
        },
    )


class LLMTokenConsumption(BaseModel):
    prompt_tokens: int
    completion_tokens: int

    def total_tokens(self) -> int:
        return self.prompt_tokens + self.completion_tokens

    def add(self, token_consumption: LLMTokenConsumption) -> "LLMTokenConsumption":
        self.prompt_tokens += token_consumption.prompt_tokens
        self.completion_tokens += token_consumption.completion_tokens
        return self


class PaginationParameters(BaseModel):
    sort: Optional[PaginationSortMethod] = PaginationSortMethod.DESCENDING
    page_size: int = 10
    page: int = 0

    def calculate_total_pages(self, total_items_count: int) -> int:
        return math.ceil(total_items_count / self.page_size)


class PIIConfig(BaseModel):
    disabled_pii_entities: Optional[list[str]] = Field(
        description=f"Optional. List of PII entities to disable. Valid values are: {PIIEntityTypes.to_string()}",
        default=None,
    )

    confidence_threshold: Optional[float] = Field(
        description=f"Optional. Float (0, 1) indicating the level of tolerable PII to consider the rule passed or failed. Min: 0 (less confident) Max: 1 (very confident). Default: {DEFAULT_PII_RULE_CONFIDENCE_SCORE_THRESHOLD}",
        default=DEFAULT_PII_RULE_CONFIDENCE_SCORE_THRESHOLD,
        json_schema_extra={"deprecated": True},
    )

    allow_list: Optional[list[str]] = Field(
        description="Optional. List of strings to pass PII validation.",
        default=None,
    )

    @field_validator("disabled_pii_entities")
    def validate_pii_entities(cls, v: list[str] | None) -> list[str] | None:
        if v:
            entities_passed = set(v)
            entities_supported = set(PIIEntityTypes.values())
            invalid_entities = entities_passed - entities_supported
            if invalid_entities:
                raise ValueError(
                    f"The following values are not valid PII entities: {invalid_entities}",
                )

            # Fail the case where they are trying to disable all PII entity types
            if (not invalid_entities) & (
                len(entities_passed) == len(entities_supported)
            ):
                raise ValueError(
                    f"Cannot disable all supported PII entities on PIIDataRule",
                )
            return v
        else:
            return v

    @field_validator("confidence_threshold")
    def validate_confidence_threshold(cls, v: float | None) -> float | None:
        if v:
            if (v < 0) | (v > 1):
                raise ValueError(f'"confidence_threshold" must be between 0 and 1')
            return v
        else:
            return v

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "disabled_pii_entities": ["PERSON", "URL"],
                "confidence_threshold": "0.5",
                "allow_list": ["arthur.ai", "Arthur"],
            },
        },
        extra="forbid",
    )


class RegexConfig(BaseModel):
    regex_patterns: List[str] = Field(
        description="List of Regex patterns to be used for validation. Be sure to encode requests in JSON and account for escape characters.",
    )

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "regex_patterns": ["\\d{3}-\\d{2}-\\d{4}", "\\d{5}-\\d{6}-\\d{7}"],
            },
        },
        extra="forbid",
    )


class ToxicityConfig(BaseModel):
    threshold: float = Field(
        default=DEFAULT_TOXICITY_RULE_THRESHOLD,
        description=f"Optional. Float (0, 1) indicating the level of tolerable toxicity to consider the rule passed or failed. Min: 0 (no toxic language) Max: 1 (very toxic language). Default: {DEFAULT_TOXICITY_RULE_THRESHOLD}",
    )

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={"example": {"threshold": DEFAULT_TOXICITY_RULE_THRESHOLD}},
    )

    @field_validator("threshold", mode="before")
    @classmethod
    def validate_toxicity_threshold(cls, v: float | None) -> float:
        if v is None:
            return float(DEFAULT_TOXICITY_RULE_THRESHOLD)
        if (v < 0) | (v > 1):
            raise ValueError(f'"threshold" must be between 0 and 1')
        return v


class UserPermission(BaseModel):
    action: UserPermissionAction
    resource: UserPermissionResource

    def __hash__(self) -> int:
        return hash((self.action, self.resource))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, UserPermission) and self.__hash__() == other.__hash__()


class VariableTemplateValue(BaseModel):
    name: str = Field(..., description="Name of the variable")
    value: str = Field(..., description="Value of the variable")



from typing import Annotated
from uuid import UUID

import pandas as pd
from duckdb import DuckDBPyConnection
from litellm import cost_per_token

from arthur_common.aggregations.aggregator import (
    NumericAggregationFunction,
    SketchAggregationFunction,
)
from arthur_common.models.enums import ModelProblemType
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    NumericMetric,
    SketchMetric,
)
from arthur_common.models.schema_definitions import (
    SHIELD_RESPONSE_SCHEMA,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
)


class ShieldInferencePassFailCountAggregation(NumericAggregationFunction):
    METRIC_NAME = "inference_count"
    FEATURE_FLAG_NAME = "SHIELD_INFERENCE_PASS_FAIL_COUNT_AGGREGATION_SEGMENTATION"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000001")

    @staticmethod
    def display_name() -> str:
        return "Inference Pass/Fail Count"

    @staticmethod
    def description() -> str:
        return "Metric that counts the number of Shield inferences grouped by the prompt, response, and overall check results."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferencePassFailCountAggregation.METRIC_NAME,
                description=ShieldInferencePassFailCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[NumericMetric]:
        # Build SELECT clause
        select_cols = [
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000)) as ts",
            "count(*) as count",
            "result",
            "inference_prompt.result AS prompt_result",
            "inference_response.result AS response_result",
        ]

        # Build GROUP BY clause
        group_by_cols = ["ts", "result", "prompt_result", "response_result"]

        # Conditionally add conversation_id and user_id based on segmentation flag
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            select_cols.extend(["conversation_id", "user_id as user_id"])
            group_by_cols.extend(["conversation_id", "user_id"])

        query = f"""
            select {", ".join(select_cols)}
            from {dataset.dataset_table_name}
            group by {", ".join(group_by_cols)}
            order by ts desc;
        """

        results = ddb_conn.sql(query).df()

        # Build group_by_dims list
        group_by_dims = [
            "result",
            "prompt_result",
            "response_result",
        ]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])

        series = self.group_query_results_to_numeric_metrics(
            results,
            "count",
            group_by_dims,
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceRuleCountAggregation(NumericAggregationFunction):
    METRIC_NAME = "rule_count"
    FEATURE_FLAG_NAME = "SHIELD_INFERENCE_RULE_COUNT_AGGREGATION_SEGMENTATION"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000002")

    @staticmethod
    def display_name() -> str:
        return "Rule Result Count"

    @staticmethod
    def description() -> str:
        return "Metric that counts the number of Shield rule evaluations grouped by whether it was on the prompt or response, the rule type, the rule evaluation result, the rule name, and the rule id."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferenceRuleCountAggregation.METRIC_NAME,
                description=ShieldInferenceRuleCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[NumericMetric]:
        # Build CTE select columns
        prompt_cte_select = [
            "unnest(inference_prompt.prompt_rule_results) as rule",
            "'prompt' as location",
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000)) as ts",
        ]
        response_cte_select = [
            "unnest(inference_response.response_rule_results) as rule",
            "'response' as location",
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000)) as ts",
        ]

        # Build main select columns
        main_select_cols = [
            "ts",
            "count(*) as count",
            "location",
            "rule.rule_type",
            "rule.result",
            "rule.name",
            "rule.id",
        ]

        # Build group by columns
        group_by_cols = [
            "ts",
            "location",
            "rule.rule_type",
            "rule.result",
            "rule.name",
            "rule.id",
        ]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            prompt_cte_select.extend(["conversation_id", "user_id"])
            response_cte_select.extend(["conversation_id", "user_id"])
            main_select_cols.extend(["conversation_id", "user_id"])
            group_by_cols.extend(["conversation_id", "user_id"])

        query = f"""
            with unnessted_prompt_rules as (select {", ".join(prompt_cte_select)}
            from {dataset.dataset_table_name}),
            unnessted_result_rules as (select {", ".join(response_cte_select)}
            from {dataset.dataset_table_name})
            select {", ".join(main_select_cols)}
            from unnessted_prompt_rules
            group by {", ".join(group_by_cols)}
            UNION ALL
            select {", ".join(main_select_cols)}
            from unnessted_result_rules
            group by {", ".join(group_by_cols)}
            order by ts desc, location, rule.rule_type, rule.result;
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = [
            "location",
            "rule_type",
            "result",
            "name",
            "id",
        ]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])
        series = self.group_query_results_to_numeric_metrics(
            results,
            "count",
            group_by_dims,
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceHallucinationCountAggregation(NumericAggregationFunction):
    METRIC_NAME = "hallucination_count"
    FEATURE_FLAG_NAME = "SHIELD_INFERENCE_HALLUCINATION_COUNT_AGGREGATION_SEGMENTATION"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000003")

    @staticmethod
    def display_name() -> str:
        return "Hallucination Count"

    @staticmethod
    def description() -> str:
        return "Metric that counts the number of Shield hallucination evaluations that failed."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferenceHallucinationCountAggregation.METRIC_NAME,
                description=ShieldInferenceHallucinationCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[NumericMetric]:
        # Build SELECT clause
        select_cols = [
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000)) as ts",
            "count(*) as count",
        ]

        # Build GROUP BY clause
        group_by_cols = ["ts"]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            select_cols.extend(["conversation_id", "user_id"])
            group_by_cols.extend(["conversation_id", "user_id"])

        query = f"""
            select {", ".join(select_cols)}
            from {dataset.dataset_table_name}
            where length(list_filter(inference_response.response_rule_results, x -> (x.rule_type = 'ModelHallucinationRuleV2' or x.rule_type = 'ModelHallucinationRule') and x.result = 'Fail')) > 0
            group by {", ".join(group_by_cols)}
            order by ts desc;
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = []
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])
        series = self.group_query_results_to_numeric_metrics(
            results,
            "count",
            group_by_dims,
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceRuleToxicityScoreAggregation(SketchAggregationFunction):
    METRIC_NAME = "toxicity_score"
    FEATURE_FLAG_NAME = "SHIELD_INFERENCE_RULE_TOXICITY_SCORE_AGGREGATION_SEGMENTATION"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000004")

    @staticmethod
    def display_name() -> str:
        return "Toxicity Distribution"

    @staticmethod
    def description() -> str:
        return "Metric that reports a distribution (data sketch) on toxicity scores returned by the Shield toxicity rule."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferenceRuleToxicityScoreAggregation.METRIC_NAME,
                description=ShieldInferenceRuleToxicityScoreAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[SketchMetric]:
        # Build CTE select columns
        prompt_cte_select = [
            "to_timestamp(created_at / 1000) as ts",
            "unnest(inference_prompt.prompt_rule_results) as rule_results",
            "'prompt' as location",
        ]
        response_cte_select = [
            "to_timestamp(created_at / 1000) as ts",
            "unnest(inference_response.response_rule_results) as rule_results",
            "'response' as location",
        ]

        # Build main select columns
        main_select_cols = [
            "ts as timestamp",
            "rule_results.details.toxicity_score::DOUBLE as toxicity_score",
            "rule_results.result as result",
            "location",
        ]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            prompt_cte_select.extend(["conversation_id", "user_id"])
            response_cte_select.extend(["conversation_id", "user_id"])
            main_select_cols.extend(["conversation_id", "user_id"])

        query = f"""
            with unnested_prompt_results as (select {", ".join(prompt_cte_select)}
            from {dataset.dataset_table_name}),
            unnested_response_results as (select {", ".join(response_cte_select)}
            from {dataset.dataset_table_name})
            select {", ".join(main_select_cols)}
            from unnested_prompt_results
            where rule_results.details.toxicity_score IS NOT NULL
            UNION ALL
            select {", ".join(main_select_cols)}
            from unnested_response_results
            where rule_results.details.toxicity_score IS NOT NULL
            order by ts desc;
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = ["result", "location"]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])

        series = self.group_query_results_to_sketch_metrics(
            results,
            "toxicity_score",
            group_by_dims,
            "timestamp",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceRulePIIDataScoreAggregation(SketchAggregationFunction):
    METRIC_NAME = "pii_score"
    FEATURE_FLAG_NAME = "SHIELD_INFERENCE_RULE_PII_DATA_SCORE_AGGREGATION_SEGMENTATION"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000005")

    @staticmethod
    def display_name() -> str:
        return "PII Score Distribution"

    @staticmethod
    def description() -> str:
        return "Metric that reports a distribution (data sketch) on PII scores returned by the Shield PII rule."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferenceRulePIIDataScoreAggregation.METRIC_NAME,
                description=ShieldInferenceRulePIIDataScoreAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[SketchMetric]:
        # Build CTE select columns
        prompt_cte_select = [
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000)) as ts",
            "unnest(inference_prompt.prompt_rule_results) as rule_results",
            "'prompt' as location",
        ]
        response_cte_select = [
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000)) as ts",
            "unnest(inference_response.response_rule_results) as rule_results",
            "'response' as location",
        ]

        # Build unnested_entities select columns
        entities_select_cols = [
            "ts",
            "rule_results.result",
            "rule_results.rule_type",
            "location",
            "unnest(rule_results.details.pii_entities) as pii_entity",
        ]

        # Build final select columns
        final_select_cols = [
            "ts as timestamp",
            "result",
            "rule_type",
            "location",
            "TRY_CAST(pii_entity.confidence AS FLOAT) as pii_score",
            "pii_entity.entity as entity",
        ]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            prompt_cte_select.extend(["conversation_id", "user_id"])
            response_cte_select.extend(["conversation_id", "user_id"])
            entities_select_cols.extend(["conversation_id", "user_id"])
            final_select_cols.extend(["conversation_id", "user_id"])

        query = f"""
            with unnested_prompt_results as (select {", ".join(prompt_cte_select)}
            from {dataset.dataset_table_name}),
            unnested_response_results as (select {", ".join(response_cte_select)}
            from {dataset.dataset_table_name}),
            unnested_entites as (select {", ".join(entities_select_cols)}
            from unnested_response_results
            where rule_results.rule_type = 'PIIDataRule'
            UNION ALL
            select {", ".join(entities_select_cols)}
            from unnested_prompt_results
            where rule_results.rule_type = 'PIIDataRule')
            select {", ".join(final_select_cols)}
            from unnested_entites
            order by ts desc;
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = ["result", "location", "entity"]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])

        series = self.group_query_results_to_sketch_metrics(
            results,
            "pii_score",
            group_by_dims,
            "timestamp",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceRuleClaimCountAggregation(SketchAggregationFunction):
    METRIC_NAME = "claim_count"
    FEATURE_FLAG_NAME = "SHIELD_INFERENCE_RULE_CLAIM_COUNT_AGGREGATION_SEGMENTATION"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000006")

    @staticmethod
    def display_name() -> str:
        return "Claim Count Distribution - All Claims"

    @staticmethod
    def description() -> str:
        return "Metric that reports a distribution (data sketch) on over the number of claims identified by the Shield hallucination rule."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferenceRuleClaimCountAggregation.METRIC_NAME,
                description=ShieldInferenceRuleClaimCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[SketchMetric]:
        # Build CTE select columns
        cte_select = [
            "to_timestamp(created_at / 1000) as ts",
            "unnest(inference_response.response_rule_results) as rule_results",
        ]

        # Build main select columns
        main_select_cols = [
            "ts as timestamp",
            "length(rule_results.details.claims) as num_claims",
            "rule_results.result as result",
        ]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            cte_select.extend(["conversation_id", "user_id"])
            main_select_cols.extend(["conversation_id", "user_id"])

        query = f"""
            with unnested_results as (select {", ".join(cte_select)}
            from {dataset.dataset_table_name})
            select {", ".join(main_select_cols)}
            from unnested_results
            where rule_results.rule_type = 'ModelHallucinationRuleV2'
            and rule_results.result != 'Skipped'
            order by ts desc;
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = ["result"]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])

        series = self.group_query_results_to_sketch_metrics(
            results,
            "num_claims",
            group_by_dims,
            "timestamp",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceRuleClaimPassCountAggregation(SketchAggregationFunction):
    METRIC_NAME = "claim_valid_count"
    FEATURE_FLAG_NAME = (
        "SHIELD_INFERENCE_RULE_CLAIM_PASS_COUNT_AGGREGATION_SEGMENTATION"
    )

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000007")

    @staticmethod
    def display_name() -> str:
        return "Claim Count Distribution - Valid Claims"

    @staticmethod
    def description() -> str:
        return "Metric that reports a distribution (data sketch) on the number of valid claims determined by the Shield hallucination rule."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferenceRuleClaimPassCountAggregation.METRIC_NAME,
                description=ShieldInferenceRuleClaimPassCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[SketchMetric]:
        # Build CTE select columns
        cte_select = [
            "to_timestamp(created_at / 1000) as ts",
            "unnest(inference_response.response_rule_results) as rule_results",
        ]

        # Build main select columns
        main_select_cols = [
            "ts as timestamp",
            "length(list_filter(rule_results.details.claims, x -> x.valid)) as num_valid_claims",
            "rule_results.result as result",
        ]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            cte_select.extend(["conversation_id", "user_id"])
            main_select_cols.extend(["conversation_id", "user_id"])

        query = f"""
            with unnested_results as (select {", ".join(cte_select)}
            from {dataset.dataset_table_name})
            select {", ".join(main_select_cols)}
            from unnested_results
            where rule_results.rule_type = 'ModelHallucinationRuleV2'
            and rule_results.result != 'Skipped'
            order by ts desc;
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = ["result"]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])

        series = self.group_query_results_to_sketch_metrics(
            results,
            "num_valid_claims",
            group_by_dims,
            "timestamp",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceRuleClaimFailCountAggregation(SketchAggregationFunction):
    METRIC_NAME = "claim_invalid_count"
    FEATURE_FLAG_NAME = (
        "SHIELD_INFERENCE_RULE_CLAIM_FAIL_COUNT_AGGREGATION_SEGMENTATION"
    )

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000008")

    @staticmethod
    def display_name() -> str:
        return "Claim Count Distribution - Invalid Claims"

    @staticmethod
    def description() -> str:
        return "Metric that reports a distribution (data sketch) on the number of invalid claims determined by the Shield hallucination rule."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferenceRuleClaimFailCountAggregation.METRIC_NAME,
                description=ShieldInferenceRuleClaimFailCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[SketchMetric]:
        # Build CTE select columns
        cte_select = [
            "to_timestamp(created_at / 1000) as ts",
            "unnest(inference_response.response_rule_results) as rule_results",
        ]

        # Build main select columns
        main_select_cols = [
            "ts as timestamp",
            "length(list_filter(rule_results.details.claims, x -> not x.valid)) as num_failed_claims",
            "rule_results.result as result",
        ]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            cte_select.extend(["conversation_id", "user_id"])
            main_select_cols.extend(["conversation_id", "user_id"])

        query = f"""
            with unnested_results as (select {", ".join(cte_select)}
            from {dataset.dataset_table_name})
            select {", ".join(main_select_cols)}
            from unnested_results
            where rule_results.rule_type = 'ModelHallucinationRuleV2'
            and rule_results.result != 'Skipped'
            order by ts desc;
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = ["result"]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])

        series = self.group_query_results_to_sketch_metrics(
            results,
            "num_failed_claims",
            group_by_dims,
            "timestamp",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceRuleLatencyAggregation(SketchAggregationFunction):
    METRIC_NAME = "rule_latency"
    FEATURE_FLAG_NAME = "SHIELD_INFERENCE_RULE_LATENCY_AGGREGATION_SEGMENTATION"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000009")

    @staticmethod
    def display_name() -> str:
        return "Rule Latency Distribution"

    @staticmethod
    def description() -> str:
        return "Metric that reports a distribution (data sketch) on the latency of Shield rule evaluations. Dimensions are the rule result, rule type, and whether the rule was applicable to a prompt or response."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ShieldInferenceRuleLatencyAggregation.METRIC_NAME,
                description=ShieldInferenceRuleLatencyAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[SketchMetric]:
        # Build CTE select columns
        prompt_cte_select = [
            "unnest(inference_prompt.prompt_rule_results) as rule",
            "'prompt' as location",
            "to_timestamp(created_at / 1000) as ts",
        ]
        response_cte_select = [
            "unnest(inference_response.response_rule_results) as rule",
            "'response' as location",
            "to_timestamp(created_at / 1000) as ts",
        ]

        # Build main select columns
        main_select_cols = [
            "ts",
            "location",
            "rule.rule_type",
            "rule.result",
            "rule.latency_ms",
        ]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            prompt_cte_select.extend(["conversation_id", "user_id"])
            response_cte_select.extend(["conversation_id", "user_id"])
            main_select_cols.extend(["conversation_id", "user_id"])

        query = f"""
            with unnested_prompt_rules as (select {", ".join(prompt_cte_select)}
            from {dataset.dataset_table_name}),
            unnested_response_rules as (select {", ".join(response_cte_select)}
            from {dataset.dataset_table_name})
            select {", ".join(main_select_cols)}
            from unnested_prompt_rules
            UNION ALL
            select {", ".join(main_select_cols)}
            from unnested_response_rules
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = ["result", "rule_type", "location"]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])

        series = self.group_query_results_to_sketch_metrics(
            results,
            "latency_ms",
            group_by_dims,
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class ShieldInferenceTokenCountAggregation(NumericAggregationFunction):
    METRIC_NAME = "token_count"
    FEATURE_FLAG_NAME = "SHIELD_INFERENCE_TOKEN_COUNT_AGGREGATION_SEGMENTATION"
    SUPPORTED_MODELS = [
        "gpt-4o",
        "gpt-4o-mini",
        "gpt-3.5-turbo",
        "o1-mini",
        "deepseek/deepseek-chat",
        "claude-3-5-sonnet-20241022",
        "gemini/gemini-1.5-pro",
        "meta.llama3-1-8b-instruct-v1:0",
        "meta.llama3-1-70b-instruct-v1:0",
        "meta.llama3-2-11b-instruct-v1:0",
    ]

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000021")

    @staticmethod
    def display_name() -> str:
        return "Token Count"

    @staticmethod
    def description() -> str:
        return "Metric that reports the number of tokens in the Shield response and prompt schemas, and their estimated cost."

    @staticmethod
    def _series_name_from_model_name(model_name: str) -> str:
        """Calculates name of reported series based on the model name considered."""
        return f"token_cost.{model_name}"

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        base_token_count_agg = BaseReportedAggregation(
            metric_name=ShieldInferenceTokenCountAggregation.METRIC_NAME,
            description=f"Metric that reports the number of tokens in the Shield response and prompt schemas.",
        )
        return [base_token_count_agg] + [
            BaseReportedAggregation(
                metric_name=ShieldInferenceTokenCountAggregation._series_name_from_model_name(
                    model_name,
                ),
                description=f"Metric that reports the estimated cost for the {model_name} model of the tokens in the Shield response and prompt schemas.",
            )
            for model_name in ShieldInferenceTokenCountAggregation.SUPPORTED_MODELS
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The task inference dataset sourced from Arthur Shield.",
                model_problem_type=ModelProblemType.ARTHUR_SHIELD,
            ),
        ],
        # This parameter exists mostly to work with the aggregation matcher such that we don't need to have any special handling for shield
        shield_response_column: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    SHIELD_RESPONSE_SCHEMA,
                ],
                friendly_name="Shield Response Column",
                description="The Shield response column from the task inference dataset.",
            ),
        ],
    ) -> list[NumericMetric]:
        # Build SELECT clause for prompt
        prompt_select_cols = [
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000)) as ts",
            "COALESCE(sum(inference_prompt.tokens), 0) as tokens",
            "'prompt' as location",
        ]

        # Build SELECT clause for response
        response_select_cols = [
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000)) as ts",
            "COALESCE(sum(inference_response.tokens), 0) as tokens",
            "'response' as location",
        ]

        # Build GROUP BY clause
        group_by_cols = [
            "time_bucket(INTERVAL '5 minutes', to_timestamp(created_at / 1000))",
            "location",
        ]

        # Conditionally add conversation_id and user_id
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            prompt_select_cols.extend(["conversation_id", "user_id"])
            response_select_cols.extend(["conversation_id", "user_id"])
            group_by_cols.extend(["conversation_id", "user_id"])

        query = f"""
            select {", ".join(prompt_select_cols)}
            from {dataset.dataset_table_name}
            group by {", ".join(group_by_cols)}
            UNION ALL
            select {", ".join(response_select_cols)}
            from {dataset.dataset_table_name}
            group by {", ".join(group_by_cols)};
        """

        results = ddb_conn.sql(query).df()

        group_by_dims = ["location"]
        if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
            group_by_dims.extend(["conversation_id", "user_id"])

        series = self.group_query_results_to_numeric_metrics(
            results,
            "tokens",
            group_by_dims,
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        resp = [metric]

        # Compute Cost for each model
        for model in self.SUPPORTED_MODELS:
            try:
                # Use litellm's cost_per_token for cost calculation
                # For each row, set prompt_tokens or completion_tokens based on location
                cost_values = []
                for tokens, location in zip(results["tokens"], results["location"]):
                    if location == "prompt":
                        prompt_cost, _ = cost_per_token(
                            model=model,
                            prompt_tokens=int(tokens),
                            completion_tokens=0,
                        )
                        cost_values.append(prompt_cost)
                    else:  # response
                        _, completion_cost = cost_per_token(
                            model=model,
                            prompt_tokens=0,
                            completion_tokens=int(tokens),
                        )
                        cost_values.append(completion_cost)
            except Exception:
                # Skip models not supported by litellm
                continue

            model_df_dict = {
                "ts": results["ts"],
                "cost": cost_values,
                "location": results["location"],
            }
            if self.is_feature_flag_enabled(self.FEATURE_FLAG_NAME):
                model_df_dict["conversation_id"] = results["conversation_id"]
                model_df_dict["user_id"] = results["user_id"]

            model_df = pd.DataFrame(model_df_dict)

            model_series = self.group_query_results_to_numeric_metrics(
                model_df,
                "cost",
                group_by_dims,
                "ts",
            )
            resp.append(
                self.series_to_metric(
                    self._series_name_from_model_name(model),
                    model_series,
                ),
            )
        return resp



import os
import re
from abc import ABC, abstractmethod
from base64 import b64encode
from typing import Any, Type, Union, cast

import pandas as pd
from datasketches import kll_floats_sketch
from duckdb import DuckDBPyConnection

from arthur_common.models.metrics import *


class AggregationFunction(ABC):
    FEATURE_FLAG_NAME: str | None = None

    @staticmethod
    @abstractmethod
    def id() -> UUID:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def display_name() -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def description() -> str:
        raise NotImplementedError

    @abstractmethod
    def aggregation_type(self) -> Type[SketchMetric] | Type[NumericMetric]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        """Returns the list of aggregations reported by the aggregate function."""
        raise NotImplementedError

    @staticmethod
    def get_innermost_segmentation_columns(segmentation_cols: list[str]) -> list[str]:
        """
        Extracts the innermost column name for nested segmentation columns or
        returns the top-level column name for non-nested segmentation columns.
        """
        for i, col in enumerate(segmentation_cols):
            # extract the innermost column for escaped column names (e.g. '"nested.col"."name"')
            # otherwise return the name since it's a top-level column
            if col.startswith('"') and col.endswith('"'):
                identifier = col[1:-1]
                identifier_split_in_struct_fields = re.split(r'"\."', identifier)

                # For nested columns, take just the innermost field name
                # Otherwise for top-level columns, take the whole name
                if len(identifier_split_in_struct_fields) > 1:
                    innermost_field = identifier_split_in_struct_fields[-1]
                    segmentation_cols[i] = innermost_field.replace('""', '"')
                else:
                    segmentation_cols[i] = identifier.replace('""', '"')
            else:
                segmentation_cols[i] = col

        return segmentation_cols

    @abstractmethod
    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        *args: Any,
        **kwargs: Any,
    ) -> Union[list[SketchMetric], list[NumericMetric]]:
        raise NotImplementedError

    @staticmethod
    def string_to_dimension(name: str, value: str | None) -> Dimension:
        if value is None:
            value = "null"
        return Dimension(name=name, value=str(value))

    def is_feature_flag_enabled(self, feature_flag_name: str) -> bool:
        if feature_flag_name is None:
            value = os.getenv(self.FEATURE_FLAG_NAME, "false")
        else:
            value = os.getenv(feature_flag_name, "false")
        return value.lower() in ("true", "1", "yes")


class NumericAggregationFunction(AggregationFunction, ABC):
    def aggregation_type(self) -> Type[NumericMetric]:
        return NumericMetric

    @abstractmethod
    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        *args: Any,
        **kwargs: Any,
    ) -> list[NumericMetric]:
        raise NotImplementedError

    @staticmethod
    def group_query_results_to_numeric_metrics(
        data: pd.DataFrame,
        value_col: str,
        dim_columns: list[str],
        timestamp_col: str,
    ) -> list[NumericTimeSeries]:
        """
        Convert a grouped dataframe with repeated dimensions to internal numeric metric definition.

        At a high level, the query results are already grouped, however,
        the order isn't guaranteed that groups are sequential (this requires an explicit ORDER BY on the source query.)
        What this function does is group by the indicated dimensions list, and from each group extract the dimension values once.
        From there, iterate over the group turning each data point to a *Point. At the end, this single instance of the group metrics
        and the list of points (values) are merged to one *TimeSeries
        """
        if not dim_columns:
            return [
                NumericAggregationFunction._dimensionless_query_results_to_numeric_metrics(
                    data,
                    value_col,
                    timestamp_col,
                ),
            ]

        # get innermost column name for nested segmentation columns
        dim_columns = AggregationFunction.get_innermost_segmentation_columns(
            dim_columns,
        )

        calculated_metrics: list[NumericTimeSeries] = []
        # make sure dropna is False or rows with "null" as a dimension value will be dropped
        groups = data.groupby(dim_columns, dropna=False)
        for _, group in groups:
            dimensions: list[Dimension] = []
            # Get the first row of the group to determine the group level dimensions
            dims_row = group.iloc[0]
            for dim in dim_columns:
                d = AggregationFunction.string_to_dimension(
                    name=dim,
                    value=dims_row[dim],
                )
                dimensions.append(d)

            values: list[NumericPoint] = []
            for _, row in group.iterrows():
                # Skip NaN values
                if pd.notna(row[value_col]):
                    values.append(
                        NumericPoint(
                            timestamp=row[timestamp_col], value=row[value_col]
                        ),
                    )
            # Only add the series if it has values
            if values:
                calculated_metrics.append(
                    NumericTimeSeries(values=values, dimensions=dimensions),
                )

        return calculated_metrics

    @staticmethod
    def _dimensionless_query_results_to_numeric_metrics(
        data: pd.DataFrame,
        value_col: str,
        timestamp_col: str,
    ) -> NumericTimeSeries:
        """
        Convert a dimensionless time / value series to internal numeric metric definition.
        """
        values: list[NumericPoint] = []
        for _, row in data.iterrows():
            # Skip NaN values
            if pd.notna(row[value_col]):
                values.append(
                    NumericPoint(timestamp=row[timestamp_col], value=row[value_col]),
                )
        return NumericTimeSeries(values=values, dimensions=[])

    @staticmethod
    def series_to_metric(
        metric_name: str,
        series: list[NumericTimeSeries],
    ) -> NumericMetric:
        return NumericMetric(name=metric_name, numeric_series=series)


class SketchAggregationFunction(AggregationFunction, ABC):
    def aggregation_type(self) -> Type[SketchMetric]:
        return SketchMetric

    @abstractmethod
    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        *args: Any,
        **kwargs: Any,
    ) -> list[SketchMetric]:
        raise NotImplementedError

    @staticmethod
    def group_query_results_to_sketch_metrics(
        data: pd.DataFrame,
        value_col: str,
        dim_columns: list[str],
        timestamp_col: str,
    ) -> list[SketchTimeSeries]:
        """
        Convert a grouped dataframe with repeated dimensions to internal sketch metric definition.

        For sketch data, what we're doing is grouping the raw row data into the dimensions we care about.
        Within each group, we extract the dimensions once. Within this single dimension group,
        we group the data into 5min intervals. Within each interval, the data point we care to sketch is added to the sketch.

        """

        calculated_metrics: list[SketchTimeSeries] = []

        # get innermost column name for nested segmentation columns
        dim_columns = AggregationFunction.get_innermost_segmentation_columns(
            dim_columns,
        )

        if dim_columns:
            # make sure dropna is False or rows with "null" as a dimension value will be dropped
            # call _group_to_series for each grouped DF
            groups = data.groupby(dim_columns, dropna=False)
            for _, group in groups:
                calculated_metrics.append(
                    SketchAggregationFunction._group_to_series(
                        group,
                        timestamp_col,
                        dim_columns,
                        value_col,
                    ),
                )
        else:
            calculated_metrics.append(
                SketchAggregationFunction._group_to_series(
                    data,
                    timestamp_col,
                    dim_columns,
                    value_col,
                ),
            )

        return calculated_metrics

    @staticmethod
    def _group_to_series(
        group: pd.DataFrame,
        timestamp_col: str,
        dim_columns: list[str],
        value_col: str,
    ) -> SketchTimeSeries:
        def to_sketch(col: pd.Series) -> Optional[kll_floats_sketch]:
            if not len(col):
                return None
            s = kll_floats_sketch()
            for v in col.values:
                s.update(v)
            return s

        dimensions: list[Dimension] = []
        if dim_columns:
            # Get the first row of the group to determine the group level dimensions
            dims_row = group.iloc[0]
            for dim in dim_columns:
                d = AggregationFunction.string_to_dimension(
                    name=dim, value=dims_row[dim]
                )
                dimensions.append(d)

        values: list[SketchPoint] = []

        # Group query results into 5min buckets
        group[timestamp_col] = pd.to_datetime(group[timestamp_col])
        group.set_index(timestamp_col, inplace=True)
        # make sure dropna is False or rows with "null" as a dimension value will be dropped
        time_bucketed_groups = group.groupby(pd.Grouper(freq="5min"), dropna=False)

        for group_timestamp, time_bucket_group in time_bucketed_groups:
            # Don't generate metrics on empty buckets
            if time_bucket_group.empty:
                continue
            sketch = to_sketch(time_bucket_group[value_col])
            if sketch is not None:
                values.append(
                    SketchPoint(
                        timestamp=cast(pd.Timestamp, group_timestamp),
                        value=b64encode(sketch.serialize()).decode(),
                    ),
                )

        return SketchTimeSeries(values=values, dimensions=dimensions)

    @staticmethod
    def series_to_metric(
        metric_name: str,
        series: list[SketchTimeSeries],
    ) -> SketchMetric:
        return SketchMetric(name=metric_name, sketch_series=series)



from typing import Annotated, Optional
from uuid import UUID

from duckdb import DuckDBPyConnection

from arthur_common.aggregations.aggregator import NumericAggregationFunction
from arthur_common.models.enums import ModelProblemType
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    NumericMetric,
)
from arthur_common.models.schema_definitions import (
    SEGMENTATION_ALLOWED_COLUMN_TYPES,
    DType,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
    MetricMultipleColumnParameterAnnotation,
    ScalarType,
    ScopeSchemaTag,
)


class MeanAbsoluteErrorAggregationFunction(NumericAggregationFunction):
    ABSOLUTE_ERROR_COUNT_METRIC_NAME = "absolute_error_count"
    ABSOLUTE_ERROR_SUM_METRIC_NAME = "absolute_error_sum"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-00000000000e")

    @staticmethod
    def display_name() -> str:
        return "Mean Absolute Error"

    @staticmethod
    def description() -> str:
        return "Metric that sums the absolute error of a prediction and ground truth column. It omits any rows where either the prediction or ground truth are null. It reports the count of non-null rows used in the calculation in a second metric."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=MeanAbsoluteErrorAggregationFunction.ABSOLUTE_ERROR_COUNT_METRIC_NAME,
                description="Sum of the absolute error of a prediction and ground truth column, omitting rows where either column is null.",
            ),
            BaseReportedAggregation(
                metric_name=MeanAbsoluteErrorAggregationFunction.ABSOLUTE_ERROR_SUM_METRIC_NAME,
                description=f"Count of non-null rows used in the calculation of the {MeanAbsoluteErrorAggregationFunction.ABSOLUTE_ERROR_SUM_METRIC_NAME} metric.",
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing the inference data.",
                model_problem_type=ModelProblemType.REGRESSION,
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        prediction_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.FLOAT),
                ],
                tag_hints=[ScopeSchemaTag.PREDICTION],
                friendly_name="Prediction Column",
                description="A column containing float typed prediction values.",
            ),
        ],
        ground_truth_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.FLOAT),
                ],
                tag_hints=[ScopeSchemaTag.GROUND_TRUTH],
                friendly_name="Ground Truth Column",
                description="A column containing float typed ground truth values.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        """Executed SQL with no segmentation columns:
                SELECT time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts, \
                SUM(ABS({prediction_col} - {ground_truth_col})) as ae, \
                COUNT(*) as count \
                FROM {dataset.dataset_table_name} \
                WHERE {prediction_col} IS NOT NULL \
                AND {ground_truth_col} IS NOT NULL \
                GROUP BY ts order by ts desc \
                """
        segmentation_cols = [] if not segmentation_cols else segmentation_cols

        # build query components with segmentation columns
        all_select_clause_cols = [
            f"time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts",
            f"SUM(ABS({prediction_col} - {ground_truth_col})) as ae",
            f"COUNT(*) as count",
        ] + segmentation_cols
        all_group_by_cols = ["ts"] + segmentation_cols

        # build query
        mae_query = f"""
            SELECT {", ".join(all_select_clause_cols)}
            FROM {dataset.dataset_table_name}
            WHERE {prediction_col} IS NOT NULL
                  AND {ground_truth_col} IS NOT NULL
            GROUP BY {", ".join(all_group_by_cols)} order by ts desc
        """

        results = ddb_conn.sql(mae_query).df()

        count_series = self.group_query_results_to_numeric_metrics(
            results,
            "count",
            segmentation_cols,
            "ts",
        )
        absolute_error_series = self.group_query_results_to_numeric_metrics(
            results,
            "ae",
            segmentation_cols,
            "ts",
        )

        count_metric = self.series_to_metric(
            self.ABSOLUTE_ERROR_COUNT_METRIC_NAME,
            count_series,
        )
        absolute_error_metric = self.series_to_metric(
            self.ABSOLUTE_ERROR_SUM_METRIC_NAME,
            absolute_error_series,
        )

        return [count_metric, absolute_error_metric]



from typing import Annotated, Optional
from uuid import UUID

from duckdb import DuckDBPyConnection

from arthur_common.aggregations.aggregator import SketchAggregationFunction
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    SketchMetric,
)
from arthur_common.models.schema_definitions import (
    SEGMENTATION_ALLOWED_COLUMN_TYPES,
    DType,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
    MetricMultipleColumnParameterAnnotation,
    ScalarType,
    ScopeSchemaTag,
)
from arthur_common.tools.duckdb_data_loader import (
    escape_str_literal,
    unescape_identifier,
)


class NumericSketchAggregationFunction(SketchAggregationFunction):
    METRIC_NAME = "numeric_sketch"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-00000000000d")

    @staticmethod
    def display_name() -> str:
        return "Numeric Distribution"

    @staticmethod
    def description() -> str:
        return (
            "Metric that calculates a distribution (data sketch) on a numeric column."
        )

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=NumericSketchAggregationFunction.METRIC_NAME,
                description=NumericSketchAggregationFunction.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing the numeric data.",
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        numeric_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.INT),
                    ScalarType(dtype=DType.FLOAT),
                ],
                tag_hints=[ScopeSchemaTag.CONTINUOUS],
                friendly_name="Numeric Column",
                description="A column containing numeric values to calculate a data sketch on.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[SketchMetric]:
        """Executed SQL with no segmentation columns:
                    select {timestamp_col} as ts, \
                       {numeric_col}, \
                       {numeric_col_name_str} as column_name \
                from {dataset.dataset_table_name} \
                where {numeric_col} is not null \
        """
        segmentation_cols = [] if not segmentation_cols else segmentation_cols
        numeric_col_name_str = escape_str_literal(unescape_identifier(numeric_col))

        # build query components with segmentation columns
        all_select_clause_cols = [
            f"{timestamp_col} as ts",
            f"{numeric_col}",
            f"{numeric_col_name_str} as column_name",
        ] + segmentation_cols
        extra_dims = ["column_name"]

        # build query
        data_query = f"""
                    select {", ".join(all_select_clause_cols)}
                    from {dataset.dataset_table_name}
                    where {numeric_col} is not null
                """

        results = ddb_conn.sql(data_query).df()

        series = self.group_query_results_to_sketch_metrics(
            results,
            unescape_identifier(numeric_col),
            segmentation_cols + extra_dims,
            "ts",
        )

        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]



import hashlib
from uuid import UUID

"""
Convert a uuid to a 12 character, all lowercase string deterministically
"""


def uuid_to_base26(uuid: str | UUID) -> str:
    # Remove hyphens from the UUID
    if isinstance(uuid, UUID):
        uuid_str = str(uuid)
    else:
        uuid_str = uuid

    no_hyphens = uuid_str.replace("-", "")

    # Convert the hex string to an integer
    num = int(no_hyphens, 16)

    # Define the alphabet for base-26
    alphabet = "abcdefghijklmnopqrstuvwxyz"
    base = len(alphabet)

    # Encode the integer into a base-26 string
    base26_str = ""
    while num > 0:
        num, rem = divmod(num, base)
        base26_str = alphabet[rem] + base26_str

    # Ensure the string is 12 characters long (pad with 'a' if necessary)
    base26_str = base26_str.rjust(12, "a")

    # If the resulting string is longer than 12 characters, take the last 12 characters
    return base26_str[-12:]


"""
Hash a string
"""


def hash_nonce(nonce: str) -> str:
    md5_hash = hashlib.md5()
    md5_hash.update(nonce.encode("utf-8"))
    return md5_hash.hexdigest()



from pydantic import BaseModel, Field


class ConnectorPaginationOptions(BaseModel):
    page: int = Field(default=1, ge=1)
    page_size: int = Field(default=25, ge=1, le=500)

    @property
    def page_params(self) -> tuple[int, int]:
        if self.page is not None:
            return (self.page, self.page_size)
        else:
            raise ValueError(
                "Pagination options must be set to return a page and page size",
            )


# connector constants
S3_CONNECTOR_ENDPOINT_FIELD = "endpoint"
AWS_CONNECTOR_REGION_FIELD = "region"
AWS_CONNECTOR_ACCESS_KEY_ID_FIELD = "access_key_id"
AWS_CONNECTOR_SECRET_ACCESS_KEY_FIELD = "secret_access_key"
AWS_CONNECTOR_ROLE_ARN_FIELD = "role_arn"
AWS_CONNECTOR_EXTERNAL_ID_FIELD = "external_id"
AWS_CONNECTOR_ROLE_DURATION_SECONDS_FIELD = "role_duration_seconds"
BUCKET_BASED_CONNECTOR_BUCKET_FIELD = "bucket"
GOOGLE_CONNECTOR_CREDENTIALS_FIELD = "credentials"
GOOGLE_CONNECTOR_PROJECT_ID_FIELD = "project_id"
GOOGLE_CONNECTOR_LOCATION_FIELD = "location"
SHIELD_CONNECTOR_API_KEY_FIELD = "api_key"
SHIELD_CONNECTOR_ENDPOINT_FIELD = "endpoint"
ODBC_CONNECTOR_HOST_FIELD = "host"
ODBC_CONNECTOR_PORT_FIELD = "port"
ODBC_CONNECTOR_DATABASE_FIELD = "database"
ODBC_CONNECTOR_USERNAME_FIELD = "username"
ODBC_CONNECTOR_PASSWORD_FIELD = "password"
ODBC_CONNECTOR_DRIVER_FIELD = "driver"
ODBC_CONNECTOR_TABLE_NAME_FIELD = "table_name"
ODBC_CONNECTOR_DIALECT_FIELD = "dialect"

# Snowflake connector constants
SNOWFLAKE_CONNECTOR_ACCOUNT_FIELD = "account"
SNOWFLAKE_CONNECTOR_SCHEMA_FIELD = "schema"
SNOWFLAKE_CONNECTOR_WAREHOUSE_FIELD = "warehouse"
SNOWFLAKE_CONNECTOR_ROLE_FIELD = "role"
SNOWFLAKE_CONNECTOR_AUTHENTICATOR_FIELD = "authenticator"
SNOWFLAKE_CONNECTOR_PRIVATE_KEY_FIELD = "private_key"
SNOWFLAKE_CONNECTOR_PRIVATE_KEY_PASSPHRASE_FIELD = "private_key_passphrase"

# Databricks connector constants
DATABRICKS_CONNECTOR_SERVER_HOSTNAME_FIELD = "server_hostname"
DATABRICKS_CONNECTOR_HTTP_PATH_FIELD = "http_path"
DATABRICKS_CONNECTOR_PAT_FIELD = "personal_access_token"
DATABRICKS_CONNECTOR_CLIENT_ID_FIELD = "client_id"
DATABRICKS_CONNECTOR_CLIENT_SECRET_FIELD = "client_secret"

# dataset (connector type dependent) constants
SHIELD_DATASET_TASK_ID_FIELD = "task_id"
BUCKET_BASED_DATASET_FILE_PREFIX_FIELD = "file_prefix"
BUCKET_BASED_DATASET_FILE_SUFFIX_FIELD = "file_suffix"
BUCKET_BASED_DATASET_FILE_TYPE_FIELD = "data_file_type"
BUCKET_BASED_DATASET_TIMESTAMP_TIME_ZONE_FIELD = "timestamp_time_zone"
BIG_QUERY_DATASET_TABLE_NAME_FIELD = "table_name"
BIG_QUERY_DATASET_DATASET_ID_FIELD = "dataset_id"
DATABRICKS_DATASET_CATALOG_FIELD = "catalog"
DATABRICKS_DATASET_SCHEMA_FIELD = "schema"

# CSV configuration field constants
CSV_DELIMITER_FIELD = "csv_delimiter"
CSV_QUOTE_CHAR_FIELD = "csv_quote_char"
CSV_ESCAPE_CHAR_FIELD = "csv_escape_char"
CSV_HAS_HEADER_FIELD = "csv_has_header"
CSV_ENCODING_FIELD = "csv_encoding"



from typing import Annotated, Optional
from uuid import UUID

from duckdb import DuckDBPyConnection

from arthur_common.aggregations.aggregator import NumericAggregationFunction
from arthur_common.models.enums import ModelProblemType
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    NumericMetric,
)
from arthur_common.models.schema_definitions import (
    SEGMENTATION_ALLOWED_COLUMN_TYPES,
    DType,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
    MetricLiteralParameterAnnotation,
    MetricMultipleColumnParameterAnnotation,
    ScalarType,
    ScopeSchemaTag,
)
from arthur_common.tools.duckdb_data_loader import (
    escape_str_literal,
    unescape_identifier,
)


class ConfusionMatrixAggregationFunction(NumericAggregationFunction):
    TRUE_POSITIVE_METRIC_NAME = "confusion_matrix_true_positive_count"
    FALSE_POSITIVE_METRIC_NAME = "confusion_matrix_false_positive_count"
    FALSE_NEGATIVE_METRIC_NAME = "confusion_matrix_false_negative_count"
    TRUE_NEGATIVE_METRIC_NAME = "confusion_matrix_true_negative_count"

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=ConfusionMatrixAggregationFunction.TRUE_POSITIVE_METRIC_NAME,
                description="Confusion matrix true positives count.",
            ),
            BaseReportedAggregation(
                metric_name=ConfusionMatrixAggregationFunction.FALSE_POSITIVE_METRIC_NAME,
                description="Confusion matrix false positives count.",
            ),
            BaseReportedAggregation(
                metric_name=ConfusionMatrixAggregationFunction.FALSE_NEGATIVE_METRIC_NAME,
                description="Confusion matrix false negatives count.",
            ),
            BaseReportedAggregation(
                metric_name=ConfusionMatrixAggregationFunction.TRUE_NEGATIVE_METRIC_NAME,
                description="Confusion matrix true negatives count.",
            ),
        ]

    def generate_confusion_matrix_metrics(
        self,
        ddb_conn: DuckDBPyConnection,
        timestamp_col: str,
        prediction_col: str,
        gt_values_col: str,
        prediction_normalization_case: str,
        gt_normalization_case: str,
        dataset: DatasetReference,
        segmentation_cols: Optional[list[str]] = None,
    ) -> list[NumericMetric]:
        """
        Generate a SQL query to compute confusion matrix metrics over time.

        Args:
            timestamp_col: Column name containing timestamps
            prediction_col: Column name containing predictions
            gt_values_col: Column name containing ground truth values
            prediction_normalization_case: SQL CASE statement for normalizing predictions to 0 / 1 / null using 'value' as the target column name
            gt_normalization_case: SQL CASE statement for normalizing ground truth values to 0 / 1 / null using 'value' as the target column name
            dataset: DatasetReference containing dataset metadata
            segmentation_cols: list of columns to segment by

        Returns:
            str: SQL query that computes confusion matrix metrics
            Without segmentation, this is the query:
                WITH normalized_data AS (
                    SELECT
                        {timestamp_col} AS timestamp,
                        {prediction_normalization_case.replace('value', prediction_col)} AS prediction,
                        {gt_normalization_case.replace('value', gt_values_col)} AS actual_value
                    FROM {dataset.dataset_table_name}
                    WHERE {timestamp_col} IS NOT NULL
                )
                SELECT
                    time_bucket(INTERVAL '5 minutes', timestamp) AS ts,
                    SUM(CASE WHEN prediction = actual_value AND actual_value = 1 THEN 1 ELSE 0 END) AS true_positive_count,
                    SUM(CASE WHEN prediction != actual_value AND actual_value = 0 THEN 1 ELSE 0 END) AS false_positive_count,
                    SUM(CASE WHEN prediction != actual_value AND actual_value = 1 THEN 1 ELSE 0 END) AS false_negative_count,
                    SUM(CASE WHEN prediction = actual_value AND actual_value = 0 THEN 1 ELSE 0 END) AS true_negative_count,
                    {unescaped_prediction_col_name} as prediction_column_name
                FROM normalized_data
                GROUP BY ts
                ORDER BY ts
        """
        segmentation_cols = [] if not segmentation_cols else segmentation_cols
        unescaped_prediction_col_name = escape_str_literal(
            unescape_identifier(prediction_col),
        )

        # build query components with segmentation columns
        first_subquery_select_cols = [
            f"{timestamp_col} AS timestamp",
            f"{prediction_normalization_case.replace('value', prediction_col)} AS prediction",
            f"{gt_normalization_case.replace('value', gt_values_col)} AS actual_value",
        ] + segmentation_cols
        second_subquery_select_cols = [
            "time_bucket(INTERVAL '5 minutes', timestamp) AS ts",
            "SUM(CASE WHEN prediction = actual_value AND actual_value = 1 THEN 1 ELSE 0 END) AS true_positive_count",
            "SUM(CASE WHEN prediction != actual_value AND actual_value = 0 THEN 1 ELSE 0 END) AS false_positive_count",
            "SUM(CASE WHEN prediction != actual_value AND actual_value = 1 THEN 1 ELSE 0 END) AS false_negative_count",
            "SUM(CASE WHEN prediction = actual_value AND actual_value = 0 THEN 1 ELSE 0 END) AS true_negative_count",
            f"{unescaped_prediction_col_name} as prediction_column_name",
        ] + segmentation_cols
        second_subquery_group_by_cols = ["ts"] + segmentation_cols
        extra_dims = ["prediction_column_name"]

        # build query
        confusion_matrix_query = f"""
                WITH normalized_data AS (
                    SELECT {", ".join(first_subquery_select_cols)}
                    FROM {dataset.dataset_table_name}
                    WHERE {timestamp_col} IS NOT NULL
                )
                SELECT {", ".join(second_subquery_select_cols)}
                FROM normalized_data
                GROUP BY {", ".join(second_subquery_group_by_cols)}
                ORDER BY ts
            """

        results = ddb_conn.sql(confusion_matrix_query).df()

        tp = self.group_query_results_to_numeric_metrics(
            results,
            "true_positive_count",
            dim_columns=segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        fp = self.group_query_results_to_numeric_metrics(
            results,
            "false_positive_count",
            dim_columns=segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        fn = self.group_query_results_to_numeric_metrics(
            results,
            "false_negative_count",
            dim_columns=segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        tn = self.group_query_results_to_numeric_metrics(
            results,
            "true_negative_count",
            dim_columns=segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        tp_metric = self.series_to_metric(self.TRUE_POSITIVE_METRIC_NAME, tp)
        fp_metric = self.series_to_metric(self.FALSE_POSITIVE_METRIC_NAME, fp)
        fn_metric = self.series_to_metric(self.FALSE_NEGATIVE_METRIC_NAME, fn)
        tn_metric = self.series_to_metric(self.TRUE_NEGATIVE_METRIC_NAME, tn)
        return [tp_metric, fp_metric, fn_metric, tn_metric]


class BinaryClassifierIntBoolConfusionMatrixAggregationFunction(
    ConfusionMatrixAggregationFunction,
):
    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-00000000001c")

    @staticmethod
    def display_name() -> str:
        return "Binary Classification Confusion Matrix - Int/Bool Prediction"

    @staticmethod
    def description() -> str:
        return "Aggregation that takes in boolean or integer prediction and ground truth values and calculates the confusion matrix (True Positives, False Positives, False Negatives, True Negatives) for a binary set of predictions and values."

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing the prediction and ground truth values.",
                model_problem_type=ModelProblemType.BINARY_CLASSIFICATION,
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        prediction_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.BOOL),
                    ScalarType(dtype=DType.INT),
                ],
                tag_hints=[ScopeSchemaTag.PREDICTION],
                friendly_name="Prediction Column",
                description="A column containing boolean or integer prediction values.",
            ),
        ],
        gt_values_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.BOOL),
                    ScalarType(dtype=DType.INT),
                ],
                tag_hints=[ScopeSchemaTag.GROUND_TRUTH],
                friendly_name="Ground Truth Column",
                description="A column containing boolean or integer ground truth values.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        segmentation_cols = [] if not segmentation_cols else segmentation_cols
        # Get the type of prediction column
        type_query = f"SELECT typeof({prediction_col}) as col_type FROM {dataset.dataset_table_name} LIMIT 1"
        res = ddb_conn.sql(type_query).fetchone()
        # As long as this column exists, we should be able to get the type. This is here to make mypy happy.
        if not res:
            raise ValueError(f"No results found for type query: {type_query}")
        col_type = res[0].lower()

        match col_type:
            case "boolean":
                normalization_case = """
                CASE
                    WHEN value THEN 1
                    ELSE 0
                END
                """
            case "integer" | "bigint":
                normalization_case = """
                CASE
                    WHEN value = 1 THEN 1
                    WHEN value = 0 THEN 0
                    ELSE NULL
                END
                """
            case _:
                raise ValueError(f"Unsupported column type: {col_type}")

        return self.generate_confusion_matrix_metrics(
            ddb_conn,
            timestamp_col,
            prediction_col,
            gt_values_col,
            normalization_case,
            normalization_case,
            dataset,
            segmentation_cols,
        )


class BinaryClassifierStringLabelConfusionMatrixAggregationFunction(
    ConfusionMatrixAggregationFunction,
):
    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-00000000001d")

    @staticmethod
    def display_name() -> str:
        return "Binary Classification Confusion Matrix - String Class Label Prediction"

    @staticmethod
    def description() -> str:
        return "Aggregation that takes in string labelled prediction and ground truth values and calculates the confusion matrix (True Positives, False Positives, False Negatives, True Negatives) for a binary set of predictions and values."

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing the prediction and ground truth values.",
                model_problem_type=ModelProblemType.BINARY_CLASSIFICATION,
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        prediction_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.STRING),
                ],
                tag_hints=[ScopeSchemaTag.PREDICTION],
                friendly_name="Prediction Column",
                description="A column containing string labelled prediction values.",
            ),
        ],
        gt_values_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.STRING),
                ],
                tag_hints=[ScopeSchemaTag.GROUND_TRUTH],
                friendly_name="Ground Truth Column",
                description="A column containing string labelled ground truth values.",
            ),
        ],
        true_label: Annotated[
            str,
            MetricLiteralParameterAnnotation(
                parameter_dtype=DType.STRING,
                friendly_name="True Label",
                description="The label indicating a positive classification to normalize to 1.",
            ),
        ],
        false_label: Annotated[
            str,
            MetricLiteralParameterAnnotation(
                parameter_dtype=DType.STRING,
                friendly_name="False Label",
                description="The label indicating a negative classification to normalize to 0.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        segmentation_cols = [] if not segmentation_cols else segmentation_cols
        normalization_case = f"""
                CASE
                    WHEN value = '{true_label}' THEN 1
                    WHEN value = '{false_label}' THEN 0
                    ELSE NULL
                END
                """
        return self.generate_confusion_matrix_metrics(
            ddb_conn,
            timestamp_col,
            prediction_col,
            gt_values_col,
            normalization_case,
            normalization_case,
            dataset,
            segmentation_cols,
        )


class BinaryClassifierProbabilityThresholdConfusionMatrixAggregationFunction(
    ConfusionMatrixAggregationFunction,
):
    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-00000000001e")

    @staticmethod
    def display_name() -> str:
        return "Binary Classification Confusion Matrix - Probability Threshold"

    @staticmethod
    def description() -> str:
        return "Aggregation that takes in a float prediction column, a ground truth values column, and a probability threshold and calculates the confusion matrix (True Positives, False Positives, False Negatives, True Negatives) for a binary set of predictions and values where the predictions are calculated using the probability threshold."

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing the prediction and ground truth values.",
                model_problem_type=ModelProblemType.BINARY_CLASSIFICATION,
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        prediction_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.FLOAT),
                ],
                tag_hints=[ScopeSchemaTag.PREDICTION],
                friendly_name="Prediction Column",
                description="A column containing float prediction values.",
            ),
        ],
        gt_values_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.BOOL),
                    ScalarType(dtype=DType.INT),
                ],
                tag_hints=[ScopeSchemaTag.GROUND_TRUTH],
                friendly_name="Ground Truth Column",
                description="A column containing boolean or integer ground truth values.",
            ),
        ],
        threshold: Annotated[
            float,
            MetricLiteralParameterAnnotation(
                parameter_dtype=DType.FLOAT,
                friendly_name="Threshold",
                description="The threshold to classify predictions to 0 or 1.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        prediction_normalization_case = f"""
                CASE
                    WHEN value >= {threshold} THEN 1
                    WHEN value < {threshold} THEN 0
                    ELSE NULL
                END
                """

        type_query = f"SELECT typeof({gt_values_col}) as col_type FROM {dataset.dataset_table_name} LIMIT 1"
        res = ddb_conn.sql(type_query).fetchone()
        # As long as this column exists, we should be able to get the type. This is here to make mypy happy.
        if not res:
            raise ValueError(f"No results found for type query: {type_query}")
        col_type = res[0].lower()

        match col_type:
            case "boolean":
                gt_normalization_case = """
                CASE
                    WHEN value THEN 1
                    ELSE 0
                END
                """
            case "integer" | "bigint":
                gt_normalization_case = """
                CASE
                    WHEN value = 1 THEN 1
                    WHEN value = 0 THEN 0
                    ELSE NULL
                END
                """
            case _:
                raise ValueError(f"Unsupported column type: {col_type}")

        return self.generate_confusion_matrix_metrics(
            ddb_conn=ddb_conn,
            timestamp_col=timestamp_col,
            prediction_col=prediction_col,
            gt_values_col=gt_values_col,
            prediction_normalization_case=prediction_normalization_case,
            gt_normalization_case=gt_normalization_case,
            dataset=dataset,
            segmentation_cols=segmentation_cols,
        )



import importlib.util
import inspect
import os

package_dir = os.path.dirname(__file__)

# Peter 05/08/2024: This is some code I swiped from stackoverflow that iterated through the package directory here looking at .py files
# It reads each file and imports the classes to add them to the "globals" which we can think of as importing into this namespace
# By doing that, everything is exported and ready to be read as members of this `functions` package.
# TLDR: this does what you would think `from . import *` does
# Benefit here is any file with any class is added to the "exports", so nothing needs to be done after dropping a file in here
for filename in os.listdir(package_dir):
    if filename.endswith(".py") and filename != "__init__.py":
        module_name = filename[:-3]  # Remove the .py extension to get the module name
        module_path = os.path.join(package_dir, filename)

        spec = importlib.util.spec_from_file_location(module_name, module_path)
        if not spec:
            continue
        module = importlib.util.module_from_spec(spec)
        if spec.loader:
            spec.loader.exec_module(module)
        for name, value in module.__dict__.items():
            if inspect.isclass(value) and not name.startswith("_"):
                globals()[name] = value
